import os
import typer
from fabric import Connection
from typing import Union
from mako.template import Template
from fastidius import __version__

def version_callback(value: bool):
    """Faciliates printing the --version."""
    if value:
        typer.echo(f"Fastidius {__version__}")
        raise typer.Exit()


def generate_file(filename: str, outfile: str = None, **kwargs) -> None:
    """Takes a Mako template file and renders the Mako logic into the file, in place."""
    routes_base = Template(filename=filename).render(**kwargs)
    if not outfile:
        outfile = filename
    with open(outfile, 'w') as file:
        file.write(routes_base)


def colored_echo(message: str, color='blue') -> None:
    """Wrapper on Typer's shell coloring."""
    COLORS = {
        'blue': typer.colors.BRIGHT_BLUE,
        'green': typer.colors.BRIGHT_GREEN,
        'red': typer.colors.BRIGHT_RED
    }
    typer.echo(typer.style(message, fg=COLORS[color]))


def welcome_prompt():
    """Prints the fastidius logo and some into text."""
    typer.echo(
        """


            ______              __   _      __ _
           / ____/____ _ _____ / /_ (_)____/ /(_)____   __  __ _____
          / /_   / __ `// ___// __// // __  // // __ \ / / / // ___/
         / __/  / /_/ /(__  )/ /_ / // /_/ // // /_/ // /_/ /(__  )
        /_/     \__,_//____/ \__//_/ \__,_//_/ \____/ \__,_//____/

        """
    )
    typer.echo(
        typer.style("fastidius needs a few settings before generating your app.\n ", fg=typer.colors.GREEN, bold=True)
    )


def ip_error(ip_address: str) -> bool:
    """Check if a valid IP address has been provided."""
    if not ip_address:
        typer.echo("[ERROR] No server IP address was supplied, please include one using either --ip-address <IP ADDRESS>, "
                   "or by setting the IP_ADDRESS environment variable.")
        return True
    return False


def connect_to_server(ip_address: str, root: bool = False) -> Union[Connection, None]:
    if ip_error(ip_address):
        raise typer.Exit()

    host = 'root' if root else 'ubuntu'
    conn = Connection(
        host=f'{host}@{ip_address}',
        connect_kwargs={
            "key_filename": f"{os.getenv('HOME')}/.ssh/id_rsa.pub",
        },
        connect_timeout=6
    )

    try:
        conn.run('ls', hide='both')
    except Exception as error:
        raise typer.Exit(f'[ERROR] Connection to the remote server was unsuccessful ({error})')

    return conn



import shutil
import os

from fastidius.services.utils import generate_file


class AppCreator:

    def __init__(self, app_name: str, FILEPATH: str, backend: bool, auth: bool, models: list) -> None:
        self.app_name = app_name
        self.backend = backend
        self.auth = auth
        self.models = models
        self.FILEPATH = FILEPATH

        # Copy the app directory over from the templates.
        shutil.copytree(f'{self.FILEPATH}/app_template', self.app_name, dirs_exist_ok=True)

    def generate(self):
        generate_file(f'{self.app_name}/docker-compose.yml', backend=self.backend, app_name=self.app_name)
        generate_file(f'{self.app_name}/README.md', backend=self.backend, app_name=self.app_name)
        generate_file(
            f'{self.app_name}/.github/workflows/test_and_deploy.yml',
            app_name=self.app_name,
            host='${{ secrets.HOST }}',
            username='${{ secrets.USERNAME }}',
            port='${{ secrets.PORT }}',
            ssh_key='${{ secrets.SSHKEY }}',
        )

        self.generate_frontend()

        # TODO: either implement or remove SQL capability
        self.remove_sqlalchemy()

        if self.backend:
            self.generate_backend()

    def generate_frontend(self):
        generate_file(f'{self.app_name}/frontend/quasar.conf.js', backend=self.backend)
        generate_file(f'{self.app_name}/frontend/package.json', app_name=self.app_name, backend=self.backend)
        generate_file(f'{self.app_name}/frontend/src/layouts/MainLayout.vue', app_name=self.app_name, auth=self.auth)
        generate_file(f'{self.app_name}/frontend/src/pages/Index.vue', app_name=self.app_name, auth=self.auth)
        generate_file(f'{self.app_name}/frontend/src/pages/Dashboard.vue', auth=self.auth)
        generate_file(f'{self.app_name}/frontend/src/store/index.js', auth=self.auth)
        generate_file(f'{self.app_name}/frontend/src/router/routes.js', auth=self.auth)
        generate_file(f'{self.app_name}/frontend/src/router/index.js', auth=self.auth)

    def generate_backend(self):
        generate_file(f'{self.app_name}/backend/main.py', auth=self.auth, alembic=True)
        generate_file(f'{self.app_name}/backend/db.py', auth=self.auth)
        generate_file(f'{self.app_name}/backend/core/settings/prod.py', app_name=self.app_name)
        generate_file(f'{self.app_name}/backend/core/settings/dev.py', app_name=self.app_name)
        generate_file(f'{self.app_name}/backend/core/settings/test.py', app_name=self.app_name)
        generate_file(f'{self.app_name}/frontend/src/router/routes.js', auth=self.auth)
        generate_file(f'{self.app_name}/frontend/src/router/index.js', auth=self.auth)
        generate_file(f'{self.app_name}/frontend/src/store/index.js', auth=self.auth)
        generate_file(f'{self.app_name}/backend/requirements.txt', app_name=self.app_name, auth=self.auth, backend=self.backend)

    def remove_backend(self):
        os.remove(f'{self.app_name}/frontend/src/boot/axios.js')

    def remove_auth(self):
        os.remove(f'{self.app_name}/backend/models/user.py')
        os.remove(f'{self.app_name}/backend/core/auth.py')
        os.remove(f'{self.app_name}/backend/api/endpoints/user_endpoints.py')
        os.remove(f'{self.app_name}/frontend/src/components/LoginForm.vue')
        os.remove(f'{self.app_name}/frontend/src/pages/Login.vue')
        os.remove(f'{self.app_name}/frontend/src/pages/Register.vue')
        os.remove(f'{self.app_name}/frontend/src/pages/RegisteredSuccessfully.vue')

    def remove_sqlalchemy(self):
        os.remove(f'{self.app_name}/backend/core/auth-sqlalchemy.py')
        os.remove(f'{self.app_name}/backend/db-sqlalchemy.py')
        os.remove(f'{self.app_name}/backend/main-sqlalchemy.py')



from typing import AsyncGenerator

% if auth:
from fastapi import Depends
from fastapi_users.db import SQLAlchemyBaseUserTable, SQLAlchemyUserDatabase
% endif
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
from sqlalchemy.orm import sessionmaker

% if auth:
from backend.models.user import UserDB
% endif
from backend.core.settings.config import settings

Base: DeclarativeMeta = declarative_base()


% if auth:
class UserTable(Base, SQLAlchemyBaseUserTable):
    pass
% endif


engine = create_async_engine(settings.DATABASE_URL)
async_session_maker = sessionmaker(engine, class_=AsyncSession, expire_on_commit=False)


async def create_db_and_tables():
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)


async def get_async_session() -> AsyncGenerator[AsyncSession, None]:
    async with async_session_maker() as session:
        yield session


% if auth:
async def get_user_db(session: AsyncSession = Depends(get_async_session)):
    yield SQLAlchemyUserDatabase(UserDB, session, UserTable)
% endif



import os
import sys
from functools import lru_cache
from typing import Union

from .dev import SettingsDev
from .prod import SettingsProd


environments: dict = {
    'dev': SettingsDev,
    'prod': SettingsProd,
}


@lru_cache
def get_app_settings() -> Union[SettingsDev, SettingsProd]:
    """All settings are routed through here on startup of the app or the test suite."""
    environment = os.environ.get('BASE_ENVIRONMENT')

    # If pytest was invoked, then use the test suite settings
    if environment != 'ci' and 'pytest' in sys.argv[0]:
        environment = 'test'

    if environment is None:
        raise ValueError("The BASE_ENVIRONMENT variable has not been set")

    config = environments[environment]
    return config()


settings = get_app_settings()



import json
from fabric import Connection
import requests
from base64 import b64encode
from nacl import encoding, public


class Github:
    def __init__(self, username: str, token: str, repo: str) -> None:
        self.username = username
        self.token = token
        self.repo = repo
        self.headers = {
            "Accept": "application/vnd.github.v3+json"
        }

    def upload_secret(self, secret_name: str, secret_value: str) -> int:
        """Upload a secret to Github secrets."""
        public_key, key_id = self._get_public_key()
        encrypted_secret = self.encrypt(public_key, secret_value)
        response = requests.put(
            f'https://api.github.com/repos/{self.username}/{self.repo}/actions/secrets/{secret_name}',
            data=json.dumps(
                {
                    "encrypted_value": encrypted_secret,
                    "key_id": key_id
                }
            ),
            auth=(self.username, self.token)
        )
        return response.status_code


    def setenv(path):
        with open(path, 'r') as envfile:
            env = envfile.read()



    def secrets_dict(self, conn: Connection, ip_address: str) -> dict:
        """Returns a structured dictionary that with fields ready to upload to Github secrets."""
        id_rsa = conn.run('cat /home/ubuntu/.ssh/id_rsa', hide='both')
        return {
            "HOST": ip_address,
            "PORT": "22",
            "USERNAME": "ubuntu",
            "SSHKEY": id_rsa.stdout
        }


    def _get_public_key(self) -> tuple:
        """Return the public key from Github needed to make further requests."""
        response = requests.get(f'https://api.github.com/repos/{self.username}/{self.repo}/actions/secrets/public-key', auth=(self.username, self.token))
        content = json.loads(response.content)
        return content.get('key'), content.get('key_id')


    def encrypt(self, public_key: str, secret_value: str) -> str:
        """Encrypt a Unicode string using the public key."""
        public_key = public.PublicKey(public_key.encode("utf-8"), encoding.Base64Encoder())
        sealed_box = public.SealedBox(public_key)
        encrypted = sealed_box.encrypt(secret_value.encode("utf-8"))
        return b64encode(encrypted).decode("utf-8")



import os
import sys
import typer
import subprocess

from invoke.exceptions import UnexpectedExit
from fastidius.services.create_app import AppCreator
from fastidius.services.github import Github
from fastidius.services.utils import colored_echo, connect_to_server, generate_file, ip_error, version_callback, welcome_prompt

cli = typer.Typer()

FILEPATH = f'{os.path.dirname(os.path.abspath(__file__))}'
IP_ADDRESS = os.getenv('IP_ADDRESS')



@cli.callback()
def common(ctx: typer.Context, version: bool = typer.Option(None, "--version", callback=version_callback)):
    """Facilitates printing the --version."""
    pass



@cli.command(help='Create a brand new web application.')
def create():
    """
    Create a web application. Includes a large list of prompts to guide the user through
    essential setup choices for their app.
    """
    welcome_prompt()

    app_name = typer.prompt("Please give your app a name: ", default='app')
    if os.path.isdir('app'):
        overwrite = typer.confirm(f"An app with the name '{app_name}' is already present, overwrite? ", default=True)
        if not overwrite:
            raise typer.Abort()

    backend = typer.confirm("Include a backend? ", default=True)
    if backend:
        auth = typer.confirm("Add authentication?", default=True)
        models = typer.prompt("Please specify the names of the initial database models (comma separated)", default='')
        models = [model.strip().capitalize() for model in models.split(',')]
    else:
        auth = False
        models = []

    creator = AppCreator(
        FILEPATH=FILEPATH,
        backend=backend,
        app_name=app_name,
        auth=auth,
        models=models,
    )

    if not auth:
        creator.remove_auth()

    if not backend:
        creator.remove_backend()

    typer.echo('Creating application...')
    creator.generate()

    # colored_echo(f'App creation was successful. You can now: cd {app_name}/', color='green')



@cli.command(help='For use directly after initializing a new VPS server. ')
def initialize_server(ip_address: str = typer.Option(IP_ADDRESS)):
    """Simple command to print out the necessary shell command to set up a new droplet."""
    if ip_error(ip_address):
        raise typer.Exit(code=1)
    typer.echo('[Run this command locally to set up a new Droplet]')
    colored_echo(f'ssh root@{ip_address} "bash -s" < {FILEPATH}/deploy/server_setup.sh\n')
    typer.echo('[If all went well, you should be able to ssh into the server. Do this at least '
               'once now, because the password for the newly created sudo user must be set]')
    colored_echo(f'ssh ubuntu@{ip_address}\n')



@cli.command(help='Generate a new Caddyfile and docker setup for caddy.')
def configure_caddy(ip_address: str = typer.Option(IP_ADDRESS)):
    """Talks to the server and configures the Caddy Server."""
    conn = connect_to_server(ip_address)
    try:
        conn.get('/caddy/Caddyfile', local=f'{FILEPATH}/deploy/', preserve_mode=False)
    except FileNotFoundError:
        typer.echo("No Caddyfile was found in /caddy, creating one...")
        generate_file(
            filename=f'{FILEPATH}/deploy/Caddyfile.mako',
            outfile=f'{FILEPATH}/deploy/Caddyfile',
            letsencrypt_email='example@hello.com',
            frontend_domain='example.com',
            backend_domain='api.example.com',
            app_name='app'
        )
        typer.echo('Generated a new Caddyfile into {FILEPATH}/deploy/Caddyfile')
    else:
        typer.echo('Successfully downloaded the Caddyfile from the server.')

    confirm = typer.confirm("Open the Caddyfile in vscode? ")
    if confirm:
        os.system(f'code {FILEPATH}/deploy/Caddyfile')



@cli.command(help='')
def deploy_caddy(ip_address: str = typer.Option(IP_ADDRESS)):
    conn = connect_to_server(ip_address, root=True)
    if not conn:
        raise typer.Exit('There was an issue connecting to the server.', code=1)

    try:
        conn.run('ls /caddy/', hide='both')
    except UnexpectedExit:
        conn.run('mkdir /caddy/')

    colored_echo(f"\n[WARNING] This action will overwrite /caddy/Caddyfile on the server with the "
                  "local version and send up the docker container.")
    confirm = typer.confirm(f"Are you happy with the contents of {FILEPATH}/deploy/Caddyfile? ")
    if confirm:
        conn.put( f'{FILEPATH}/deploy/Caddyfile', remote="/caddy/Caddyfile",  preserve_mode=False)
        conn.put( f'{FILEPATH}/deploy/docker-compose.yml', remote="/caddy/docker-compose.yml",  preserve_mode=False)
        typer.echo('Files uploaded. Sending up the Caddy container...')
        conn.run('cd /caddy/ && docker-compose down -v && docker-compose up --build -d', echo=True)




@cli.command(help='Set up the Github Action secrets necessary for deployment.')
def github_setup(
        github_username: str = typer.Option('', envvar='GITHUB_USERNAME'),
        github_token: str = typer.Option('',  envvar='GITHUB_TOKEN'),
        ip_address: str = typer.Option('',  envvar='IP_ADDRESS'),
        github_repo: str = typer.Option(''),
    ):
    """Sets up the Github environment by adding secrets necessary for deployment in Github Actions."""
    if not github_token or not github_username:
        raise ValueError('No github username/token found. Please set either the --github-token ' +
                         'and --github-username flags, or the GITHUB_TOKEN and GITHUB_USERNAME shell variables.')
    if not github_repo:
        github_repo = os.path.basename(os.getcwd())

    github = Github(username=github_username, token=github_token, repo=github_repo)
    conn = connect_to_server(ip_address)
    SECRETS = github.secrets_dict(conn, ip_address)
    for secret_name, secret_value in SECRETS.items():
        response_code = github.upload_secret(secret_name=secret_name, secret_value=secret_value)
        if response_code == 204:
            typer.echo(f'"{secret_name}" was successfully uploaded to Github secrets.')





if __name__ == "__main__":
    cli()



from fastapi import APIRouter, Depends
from backend.models.user import UserDB
from backend.core.auth import current_active_user

router = APIRouter()


@router.get("/authentication-check")
async def authentication_check(user: UserDB = Depends(current_active_user)):
    return {"message": f"Hello {user.email}!"}



from typing import Optional

from fastapi import Depends, Request
from fastapi_users import BaseUserManager, FastAPIUsers
from fastapi_users.authentication import (
    AuthenticationBackend,
    CookieTransport,
    JWTStrategy,
)
from fastapi_users.db import MongoDBUserDatabase

from backend.db import get_user_db
from backend.models.user import User, UserCreate, UserDB, UserUpdate

SECRET = "SECRET"


class UserManager(BaseUserManager[UserCreate, UserDB]):
    user_db_model = UserDB
    reset_password_token_secret = SECRET
    verification_token_secret = SECRET

    async def on_after_register(self, user: UserDB, request: Optional[Request] = None):
        print(f"User {user.id} has registered.")

    async def on_after_forgot_password(
        self, user: UserDB, token: str, request: Optional[Request] = None
    ):
        print(f"User {user.id} has forgot their password. Reset token: {token}")

    async def on_after_request_verify(
        self, user: UserDB, token: str, request: Optional[Request] = None
    ):
        print(f"Verification requested for user {user.id}. Verification token: {token}")


async def get_user_manager(user_db: MongoDBUserDatabase = Depends(get_user_db)):
    yield UserManager(user_db)


cookie_transport = CookieTransport(cookie_max_age=3600)


def get_jwt_strategy() -> JWTStrategy:
    return JWTStrategy(secret=SECRET, lifetime_seconds=3600)


auth_backend = AuthenticationBackend(
    name="jwt",
    transport=cookie_transport,
    get_strategy=get_jwt_strategy,
)
fastapi_users = FastAPIUsers(
    get_user_manager,
    [auth_backend],
    User,
    UserCreate,
    UserUpdate,
    UserDB,
)

current_active_user = fastapi_users.current_user(active=True)



from fastapi import Depends, FastAPI
from fastapi.middleware.cors import CORSMiddleware

% if auth:
from backend.models.user import UserDB
from backend.core.auth import auth_backend, current_active_user, fastapi_users
from backend.api.endpoints.user_endpoints import router as user_routes
% endif


app = FastAPI()

origins = [
    "http://localhost:8080",
    "http://localhost:3000",
]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


% if auth:
app.include_router(
    fastapi_users.get_auth_router(auth_backend), prefix="/auth/jwt", tags=["auth"]
)
app.include_router(fastapi_users.get_register_router(), prefix="/auth", tags=["auth"])
app.include_router(
    fastapi_users.get_reset_password_router(),
    prefix="/auth",
    tags=["auth"],
)
app.include_router(
    fastapi_users.get_verify_router(),
    prefix="/auth",
    tags=["auth"],
)
app.include_router(fastapi_users.get_users_router(), prefix="/users", tags=["users"])


@app.get("/authenticated-route")
async def authenticated_route(user: UserDB = Depends(current_active_user)):
    return {"message": f"Hello {user.email}!"}
% endif



import os
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read()

setup(
    name='fastidius',
    version=os.getenv('FASTIDIUS_VERSION'),
    author='John Kealy',
    author_email='johnckealy.dev@gmail.com',
    license='MIT',
    description='A full stack web app creator with FastAPI',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/johnckealy/fastidius',
    py_modules=['fastidius'],
    packages=find_packages(),
    package_data={
    },
    include_package_data=True,
    install_requires=[requirements],
    python_requires='>=3.9',
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    entry_points='''
        [console_scripts]
        fastidius=fastidius.cli:cli
    '''
)



from fastapi_users import models


class User(models.BaseUser):
    pass


class UserCreate(models.BaseUserCreate):
    pass


class UserUpdate(models.BaseUserUpdate):
    pass


class UserDB(User, models.BaseUserDB):
    pass



import motor.motor_asyncio
from fastapi_users.db import MongoDBUserDatabase

from backend.models.user import UserDB
from backend.core.settings.config import settings


client = motor.motor_asyncio.AsyncIOMotorClient(
    settings.DATABASE_URL, uuidRepresentation="standard"
)

db = client["database_name"]
collection = db["users"]


async def get_user_db():
    yield MongoDBUserDatabase(UserDB, collection)
